/*

Copyright (C) 1996, 1997 John W. Eaton
Copyright (C) 2006 Pascal Dupuis
This file is part of Octave.

Octave is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

Octave is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Octave; see the file COPYING.  If not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301, USA.

*/

#if defined (__GNUG__) && defined (USE_PRAGMA_INTERFACE_IMPLEMENTATION)
#pragma implementation
#endif

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>

#include "CmplxGSVD.h"
#include "f77-fcn.h"
#include "lo-error.h"

extern "C"
{
  F77_RET_T
  F77_FUNC (zggsvd, ZGGSVD)  
   (
     F77_CONST_CHAR_ARG_DECL, 	// JOBU    (input) CHARACTER*1
     F77_CONST_CHAR_ARG_DECL, 	// JOBV    (input) CHARACTER*1
     F77_CONST_CHAR_ARG_DECL, 	// JOBQ    (input) CHARACTER*1
     const int&, 		// M       (input) INTEGER
     const int&, 		// N       (input) INTEGER
     const int&,		// P       (input) INTEGER
     int &, 			// K       (output) INTEGER
     int &,			// L       (output) INTEGER
     Complex*,			// A       (input/output) COMPLEX*16 array, dimension (LDA,N)
     const int&, 		// LDA     (input) INTEGER
     Complex*, 			// B       (input/output) COMPLEX*16 array, dimension (LDB,N)
     const int&, 		// LDB     (input) INTEGER
     double*, 			// ALPHA   (output) DOUBLE PRECISION array, dimension (N)
     double*, 			// BETA    (output) DOUBLE PRECISION array, dimension (N)
     Complex*,			// U       (output) COMPLEX*16 array, dimension (LDU,M)
     const int&,		// LDU     (input) INTEGER 
     Complex*, 			// V       (output) COMPLEX*16 array, dimension (LDV,P)
     const int&,		// LDV     (input) INTEGER
     Complex*,			// Q       (output) COMPLEX*16 array, dimension (LDQ,N) 
     const int&,		// LDQ     (input) INTEGER
     Complex*, 			// WORK    (workspace) COMPLEX*16 array
     double*,			// RWORK   (workspace) DOUBLE PRECISION array
     int*,	 		// IWORK   (workspace/output) INTEGER array, dimension (N)
     int&			// INFO    (output)INTEGER
     F77_CHAR_ARG_LEN_DECL
     F77_CHAR_ARG_LEN_DECL
     );
}

ComplexMatrix
ComplexGSVD::left_singular_matrix_A (void) const
{
  if (type_computed == GSVD::sigma_only)
    {
      (*current_liboctave_error_handler)
	("dbleGSVD: U not computed because type == GSVD::sigma_only");
      return ComplexMatrix ();
    }
  else
    return left_smA;
}

ComplexMatrix
ComplexGSVD::left_singular_matrix_B (void) const
{
  if (type_computed == GSVD::sigma_only)
    {
      (*current_liboctave_error_handler)
	("dbleGSVD: V not computed because type == GSVD::sigma_only");
      return ComplexMatrix ();
    }
  else
    return left_smB;
}

ComplexMatrix
ComplexGSVD::right_singular_matrix (void) const
{
  if (type_computed == GSVD::sigma_only)
    {
      (*current_liboctave_error_handler)
	("dbleSVD: X not computed because type == GSVD::sigma_only");
      return ComplexMatrix ();
    }
  else
    return right_sm;
}

int
ComplexGSVD::init (const ComplexMatrix& a, const ComplexMatrix& b, 
		   GSVD::type gsvd_type)
{
  int info;

  int m = a.rows ();
  int n = a.cols ();
  int p = b.rows ();
  
  ComplexMatrix atmp = a;
  Complex *tmp_dataA = atmp.fortran_vec ();
  
  ComplexMatrix btmp = b;
  Complex *tmp_dataB = btmp.fortran_vec ();

  int min_mn = m < n ? m : n;

  char jobu = 'U';
  char jobv = 'V';
  char jobq = 'Q';

  int nrow_u = m;
  int nrow_v = p;
  int nrow_q = n;

  int k, l;

  switch (gsvd_type)
    {

    case GSVD::sigma_only:

      // Note:  for this case, both jobu and jobv should be 'N', but
      // there seems to be a bug in dgesvd from Lapack V2.0.  To
      // demonstrate the bug, set both jobu and jobv to 'N' and find
      // the singular values of [eye(3), eye(3)].  The result is
      // [-sqrt(2), -sqrt(2), -sqrt(2)].
      //
      // For Lapack 3.0, this problem seems to be fixed.

      jobu = 'N';
      jobv = 'N';
      jobq = 'N';
      nrow_u = nrow_v = nrow_q = 1;
      break;

    default:
      break;
    }

  type_computed = gsvd_type;

  if (! (jobu == 'N' || jobu == 'O')) {
    left_smA.resize (nrow_u, m);
  }
  
  Complex *u = left_smA.fortran_vec ();

  if (! (jobv == 'N' || jobv == 'O')) {
    left_smB.resize (nrow_v, p);
  }

  Complex *v = left_smB.fortran_vec ();

  sigmaA.resize (n, n);
  double *c_vec  = sigmaA.fortran_vec ();

  sigmaB.resize (n, n);
  double *s_vec  = sigmaB.fortran_vec ();

  if (! (jobq == 'N' || jobq == 'O')) {
    right_sm.resize (nrow_q, n);
  }
  Complex *q = right_sm.fortran_vec ();  
  
  int lwork = 3*n;
  lwork = lwork > m ? lwork : m;
  lwork = lwork > p ? lwork : p;
  Array<Complex>  work (lwork);
  Array<double>   rwork(2*n);
  Array<int>      iwork (n);

  F77_XFCN (zggsvd, ZGGSVD, (F77_CONST_CHAR_ARG2 (&jobu, 1),
			     F77_CONST_CHAR_ARG2 (&jobv, 1),
			     F77_CONST_CHAR_ARG2 (&jobq, 1),
			     m, n, p, k, l, tmp_dataA, m, 
			     tmp_dataB, p, c_vec, s_vec, u, nrow_u, 
			     v, nrow_v, q, nrow_q, work.fortran_vec (), 
			     rwork.fortran_vec (), iwork.fortran_vec (), 
			     info
			     F77_CHAR_ARG_LEN (1)
			     F77_CHAR_ARG_LEN (1)));
  
  if (f77_exception_encountered)
    (*current_liboctave_error_handler) ("unrecoverable error in zggsvd");

  return info;
}

/*
;;; Local Variables: ***
;;; mode: C++ ***
;;; End: ***
*/
